// $Id: TKqml.h,v 1.5 1997/06/04 16:37:06 ian Exp $

#define TALLOC(TYPE) (TYPE *)malloc(sizeof(TYPE))
#define KQML_PERF_EXTENSIONS

#define MAX_PARAMS 8

// threshhold above which strings are passed as references
#define MAX_STRING 16

#define ERR -1
#define OK 0
#define KQML_ERR ERR
#define KQML_OK OK
#define TKQML_ERR ERR
#define TKQML_OK OK

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define TCL_NULL ""
#define MAX_PARAM 30

#define TKQML_PTR_LEN 9
#define TKQML_PCHAR_TYPE 'c'
#define TKQML_QCHAR_TYPE 'q'
#define TKQML_POBJ_TYPE  'p'
#define TKQML_PPOBJ_TYPE 'o'
#define TKQML_UNDEF_TYPE 'u'

#define DEBUG_OUT stderr

#ifdef DEBUG
#define dprint if(TRUE)printf
#else
#define dprint if(FALSE)printf
#endif

#define TKQML_ERROR_KRIL_FORMAT      "Go away!"
#define TKQML_ERROR_NOT_A_STRING     "Go away!"
#define TKQML_ERROR_UNKNOWN_COMMAND  "ERR: Unknown command."
#define TKQML_ERROR_DEPRECATED_COMMAND "ERR: Deprecated KRIL function."
#define TKQML_ERROR_NUM_PARAMS     "ERR: Incorrect number of parameters."
#define TKQML_ERROR_CANNOT_INIT    "ERR: Cannot init."

#define TKQML_MOUNT_ERROR         "ERR: Unable to select C library routine."

#define TKQML_ENUM_TYPE_ERROR     "EER: Illegal value for enumerated type."

#define TKQML_TYPE_ERROR          "ERR: Type error."
#define TKQML_TYPE_COPY_ERROR     "ERR: Parameter type cannot be modified."
#define TKQML_TYPE_INT_ERROR      "ERR: Converting INT parameter."
#define TKQML_TYPE_UNKNOWN_ERROR  "ERR: Cannot dock; type spec. unknown."

#define TKQML_DERIV_INT_ERROR     "ERR: Cannot derive INT."
#define TKQML_DERIV_PINT_ERROR    "ERR: Cannot derive PINT."
#define TKQML_DERIV_CHAR_ERROR    "ERR: Cannot derive CHAR."
#define TKQML_DERIV_PCHAR_ERROR   "ERR: Cannot derive PCHAR."
#define TKQML_DERIV_POBJ_ERROR    "ERR: Cannot derive POBJ."
#define TKQML_DERIV_PPOBJ_ERROR   "ERR: Cannot derive PPOBJ."
#define TKQML_DERIV_UNKNOWN_ERROR "ERR: Unknown type; cannot derive."
#define TKQML_DERIV_NUL_ERROR     "ERR: NUL not a valid type for conversion."
#define TKQML_DERIV_KET_ERROR     "ERR: Cannot derive KET."
#define TKQML_DERIV_KRT_ERROR     "ERR: Cannot derive KRT."
#define TKQML_DERIV_PKET_ERROR     "ERR: Cannot derive PKET."
#define TKQML_DERIV_PKRT_ERROR     "ERR: Cannot derive PKRT."

#define TKQML_STATIC   0
#define TKQML_VOLATILE 1
#define TKQML_DYNAMIC  3

#define RETURN_CODE_TYPE 0x000000FF
#define RETURN_CODE_INST 0x0000FF00

// Flag bits for type return codes
#define NO_ACTION 0x00000000  /* do nothing                       */
#define DESTROY   0x00000200  /* recycle memory                   */
#define KINT      0x00000002  /* primitive type: integer          */
#define KCHAR     0x00000004  /*                 character        */
#define KPTR      0x00000008  /*                 pointer          */
#define KPPTR     0x00000020  /*                 pointer to pointer */
#define KSTR      0x0000000F  /*                 string           */
#define KKET      0x00000010  /* enum. type: kqml_expression_type */
#define KKRT      0x00000010  /*             kqml_reply_type      */
#define COPY      0x00000100  /* copy value back to variable      */

#define TTYPE long unsigned int

// TKqml type codes (5)
#define NUL     ( TTYPE ) 0x00000000  /* NULL */ 
#define INT     ( TTYPE ) 0x00000001  /* int   */
#define PINT    ( TTYPE ) 0x00000002  /* int*  */
#define CHAR    ( TTYPE ) 0x00000004  /* char  */
#define PCHAR   ( TTYPE ) 0x00000008  /* char* */
#define POBJ    ( TTYPE ) 0x0000000F  /* void* */
#define QCHAR   ( TTYPE ) 0x00000010  /* char** */
#define KET     ( TTYPE ) 0x00000020  /* kqml_expression_type (enum) */
#define PKET    ( TTYPE ) 0x00000040  /* ptr kqml_expression_type (enum) */
#define KRT     ( TTYPE ) 0x00000080  /* kqml_reply_type (enum) */
#define PKRT    ( TTYPE ) 0x000000F0  /* ptr kqml_reply_type (enum) */
#define PPOBJ   ( TTYPE ) 0x00000100  /* void** */

typedef int (*DType)(Tcl_Interp*, char*, TTYPE *rc, void**);

typedef struct handler_script { // extensible
  char *perf;    // performative
  char *handler; // handler script
  int  handler_is_file; // handler is a filename
} HANDLER_TABLE;

typedef struct kqml_cmd {
  char *cmd;
  int np;
  int vp;
  TTYPE t[MAX_PARAM];
  TTYPE rt; // return type
} KQML_CMDARRAY;

typedef struct ttype_cmd {
  TTYPE        type;
  DType        proc;
} TTYPE_CMDARRAY;

extern HANDLER_TABLE handler_table[];
extern const int kqml_numhand;
extern KQML_CMDARRAY kqml_cmdarray[];
extern const int kqml_numcmds;
extern TTYPE_CMDARRAY ttype_cmdarray[];
extern const int ttype_numcmds;








