//-----------------------------------------------------------------------------
// 
//      TclInterface.C: Tcl/Kqml Interface 
//      R. Scott Cost 3/27/96
//
//      Code to provide kqml for Tcl.
//      This code will eventually be used for a standalone c/kqml shell.
//      In the interim, it should be included in ttTclInterface.C.  Also,
//         1. the appropriate kqml header files must be in tt...C
//         2. Makefile must know app. kqml directores, libs, etc.
//         3. AppInit must contain an entry for kqml_command
//
// $Id: TKqmlApp.C,v 1.6 1997/06/04 16:37:31 ian Exp $
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <minmax.h>
#include <ctype.h>
#include <stdlib.h>
/* #include <strings.h> */
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <kqml.h>
#include <TKqml.h>
#include <TKqmlApp.h>

int KQML_Init (Tcl_Interp *interp)
{
  Tcl_CreateCommand (interp, "kqml", Kqml_Command,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *)NULL);
     return (TKQML_OK);
}

static int TKqml_initialize (ClientData clientData, Tcl_Interp *interp,
	int argc, char* argv[])
{
  TInterp = interp;            // establish univ. accessible pointer to interp.
  if (argc != 3)
    TKqml_Exit(interp,TKQML_ERROR_NUM_PARAMS,TKQML_STATIC,TCL_ERROR);
  if (kqml_initialize(argv[2],KQML_NOT_AN_ANS) < 0)  
    TKqml_Exit(interp,TKQML_ERROR_CANNOT_INIT,TKQML_STATIC,TCL_ERROR);
  kqml_register(master_handler,"",1);                 // async. default handler
  return(TCL_OK);
}

static int TKqml_deref (ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
    char *new_ptr, *ptr;
    if (argc != 3) {
	TKqml_Exit(interp, TKQML_ERROR_NUM_PARAMS, TKQML_STATIC, TCL_ERROR);
    }

    if ((ptr=Tcl_GetVar(interp, argv[2], TCL_GLOBAL_ONLY))) {
	if ((new_ptr = deref(ptr)) == NULL)
	    return(tkqml_err("Error dereferencing PPOBJ"));
	else {
	    Tcl_SetResult(interp, new_ptr, TCL_VOLATILE);
	    return TCL_OK;
	}
    }
    else if ((new_ptr = deref(argv[2])) == NULL)
	return(tkqml_err("Error dereferencing PPOBJ"));
    else {
	Tcl_SetResult(interp, new_ptr, TCL_VOLATILE);
	return TCL_OK;
    }
}

//-----------------------------------------------------------------------------
//
//	Kqml_Command (Dispatch)
//
//	Process a KQML command from Tcl.  Every Kril function made avail-
//      able to Tcl has a corresponding mount, a procedure which aligns
//      the parameters and invokes the Kril function.  This dispatch 
//      routine chooses the correct mount for the command, and invokes it.
//      
//      The available mounts are specified in kqml_cmdarray.
//
//      ref:  argc "kqml build_msg tell foo bob id" = 6
//      ref:  argv[0] (of above) = "kqml"
//-----------------------------------------------------------------------------

/* still needs checking on number of parameters */

static int Kqml_Command (ClientData clientData, Tcl_Interp *interp,
		  int argc, char *argv[])
{
  void *argt[MAX_PARAM];
  TTYPE return_code[MAX_PARAM];
  int rval;
  int clen;
  int i, a, b, c, d;
  
  if (argc < 2) 
    TKqml_Exit(interp,TKQML_ERROR_UNKNOWN_COMMAND,TKQML_STATIC,TCL_ERROR);
  else if (!strcasecmp (argv[1], "initialize"))
    rval=TKqml_initialize(clientData,interp,argc,argv); 
  else if (!strcasecmp (argv[1], "register"))
    rval=register_handler(clientData,interp,argc,argv); 
  else if (!strcasecmp (argv[1], "register_script"))
    rval=register_handler_as_script(clientData,interp,argc,argv); 
  else if (!strcasecmp (argv[1], "deref"))
    rval=TKqml_deref(clientData, interp, argc, argv);
  else {
    argt[argc] = (void*)NULL;
    clen = strlen(argv[1]);
    for (i = 0; i < kqml_numcmds; i++) {
      if (!strncasecmp (argv[1], kqml_cmdarray[i].cmd, clen)) {
	// Deprecated functions have a "-NA" suffix, so check length
	if (clen != strlen(kqml_cmdarray[i].cmd)) {
	  if (strstr("-NA", kqml_cmdarray[i].cmd) != NULL) {
	    TKqml_Exit(interp,TKQML_ERROR_DEPRECATED_COMMAND,
		       TKQML_STATIC,TCL_ERROR);
	    return (0);
	  }
	  else continue;
	}
	
	a=kqml_cmdarray[i].np;
	b=kqml_cmdarray[i].vp;
	c=argc-2;
	if (((d=(c-a))>=0) && (((b==d)&&(b==0)) || ((d-((d/b)*b))==0)))
	{          // num param check
	  argt[1]=(void*)argv[1];                    // this value is read only
	  rval=dock(clientData, interp, argc, argv, argt, return_code, i);
	  if (rval) return(rval);
	  rval = kril_mount(clientData, interp, argc, argt, i);
	  launch(clientData, interp, argc, argv, argt, return_code, i);
	  return(rval);
	}
      }
    } /* you've gone too far! */
    TKqml_Exit(interp,TKQML_ERROR_UNKNOWN_COMMAND,TKQML_STATIC,TCL_ERROR);
  }
  return (rval);
}

static int dock(ClientData clientData, Tcl_Interp* interp, int argc, 
		char *argv[], void *argt[], TTYPE rc[], int index) {
  int i, j, rval;
  TTYPE type;

  for(i=2; i<argc; i++) {
    if ((i-2)>=kqml_cmdarray[index].np) type=PCHAR;                   // vararg
    else type=kqml_cmdarray[index].t[i-2];
    for (j=0; j < ttype_numcmds; j++) {
      if (ttype_cmdarray[j].type==type) {                    // type identified
	if (rval=ttype_cmdarray[j].proc(interp,argv[i],&(rc[i]),&(argt[i]))) {
	  return(rval);                                   // error has occurred
	}
	else break;                    // out of forj (type), to next i (param)
      }
    }
    if (j==ttype_numcmds) return(tkqml_err(TKQML_DERIV_UNKNOWN_ERROR));
  }
  return(TCL_OK);
}

//-----------------------------------------------------------------------------
//
//	kril_mount
//
//	This is the point at which all kril functions are attached to the 
//      application.  The appropriate function is identified (by name), 
//      called, and the value is placed in the argt buffer.  Although all 
//      parameters must be type-cast from pointers, all transformations take
//      place before and after the mount.
//
//-----------------------------------------------------------------------------

static int kril_mount(ClientData clientData, Tcl_Interp *interp, int argc, 
		      void *a[], int index) { // a<-argt
  kqml_message *msg;
  static char *filler = "filler";

  a[0]=a[1]=(void*)TALLOC(int); // default return value is integer

  switch (index) {
  case 0 : 
    break;
  case 1 :
    break;
  case 2 :
    break;
  case 3 :
    break;
  case 4 : // send_message
      msg=kqml_build_msg((char*)a[3],(char*)a[4],(char*)a[5],(char*)a[6], 
			 NULL);
    if ((argc-2)>kqml_cmdarray[index].np) 
      add_fields(msg,(char **)(&a[kqml_cmdarray[index].np+2]),index);
    *(int*)a[0]=kqml_send(*(int*)a[2],msg,(kqml_message **)a[4]);
    break;
  case 5 : // send_msgv - DEPRECATED
    msg=kqml_build_msgv((char*)a[3],(char*)a[4],(char*)a[5],(char*)a[6],
			(char**)a[7],(char**)a[8]);
    *(int*)a[0]=kqml_send(*(int*)a[2],msg,(kqml_message **)a[4]);
    break;
  case 6 : // send
    *(int*)a[0]=kqml_send(*(int*)a[2],(kqml_message*)a[3],
			  (kqml_message**)a[4]);
    break;
  case 7: // deliver_msg
    /* filler values for receiver and reply-tag will be replaced on send */
    msg=kqml_build_msg((char*)a[2],(char*)a[3],filler,filler);
    if ((argc-2)>kqml_cmdarray[index].np) 
      add_fields(msg,(char **)(&a[kqml_cmdarray[index].np+2]),index);
    *(int*)a[0]=kqml_deliver(msg);
    break;
  case 8 : // deliver_msgv - DEPRECATED
    msg=kqml_build_msgv((char*)a[2],(char*)a[3],filler,filler,(char**)a[4],
			(char**)a[5]);
    *(int*)a[0]=kqml_deliver(msg);
    break;
  case 9 : // deliver
    *(int*)a[0]=kqml_deliver((kqml_message*)a[2]);
    break;
  case 10 : // build_msg
    msg=kqml_build_msg((char*)a[2],(char*)a[3],(char*)a[4],(char*)a[5],NULL);
    if ((argc-2)>kqml_cmdarray[index].np) 
      add_fields(msg,(char **)(&a[kqml_cmdarray[index].np+2]),index);
    a[0]=(void*)msg;
    break;
  case 11 : // build_msgv - DEPRECATED
    a[0]=(void*)kqml_build_msg((char*)a[2],(char*)a[3],(char*)a[4],
			       (char*)a[5],(char**)a[6],(char**)a[7]);
    break;
  case 12 : // get_performative
    a[0]=(void*)kqml_get_performative((kqml_message*)a[2]);
    break;
  case 13 : // get_field
    a[0]=(void*)kqml_get_field((char*)a[2],(kqml_message*)a[3],
			       (int*)a[4],(kqml_expression_type *)a[5]);
    break;
  case 14 : //
    *(int*)a[0]=kqml_get_field_count((kqml_message*)a[2]);
    break;
  case 15 : //
    break;
  case 16 : //
    *(int*)a[0]=kqml_put_performative((kqml_message*)a[2],(char*)a[3]);
    break;
  case 17 : //
    *(int*)a[0]=kqml_put_field((char*)a[2],(kqml_message*)a[3],(char*)a[4],
			       *(int*)a[5]);
    break;
  case 18 : //
    *(int*)a[0]= kqml_put_embedded_msg((char*)a[2],(kqml_message*)a[3],
				       (kqml_message*)a[4]);
    break;
  case 19 : //
    a[0]=(void *)kqml_build_kmsg_list((KQML_KMSG_LIST **)(&a[2]),
				      (kqml_message*)a[3]);
    break;
  case 20 : //
    kqml_free_kmsg_list((KQML_KMSG_LIST *)a[2]);
    break;
  case 21 : //
    kqml_print_kmsg_list((FILE *)a[2],(KQML_KMSG_LIST *)a[3]);
    break;
  case 22 : //
    kqml_print_msg((FILE *)a[2],(kqml_message *)a[3]);
    break;
  case 23 : //
    a[0]=(void *)kqml_sprint_msg((char*)a[2],(int*)a[3],(kqml_message*)a[4]);
    break;
  case 24 : //
    *(int*)a[0]= kqml_parse_buffer((char*)a[2],*(int*)(a[3]),(int*)a[4],
				   (kqml_message**)a[5]);
    break;
  case 25 : //
    *(int*)a[0]=kqml_parse_stream((FILE*)a[2],(kqml_message **)a[3]);
    break;
  case 26 : //
    *(int*)a[0]=kqml_parse_fd(*(int*)a[2],(kqml_message **)a[3]);
    break;
  case 27 : //
    *(int*)a[0]=kqml_timings();
    break;
  case 28 : // Foo?
    break;
  case 29 : //
    kqml_timings_start();
    break;
  case 30 : //
    kqml_timings_stop();
    break;
  case 31 : //
    *(int*)a[0]=kqml_logging();
    break;
  case 32 : //
    kqml_logging_start();
    break;
  case 33 : //
    kqml_logging_stop();
    break;
  case 34 : //
    a[0]=(void*)kqml_wrapper((KQML_REPLY_TYPE)a[2],a[3]);
    break;
  case 35 : //
    kqml_unwrapper((kqml_message*)a[2],(KQML_REPLY_TYPE *)a[3]);
    break;
  case 36 : //
    *(int*)a[0]=kqml_stricmp((char*)a[2],(char*)a[3]);
    break;
  case 37 : //
    *(int*)a[0]=kqml_strincmp((char*)a[2],(char*)a[3],*(int*)(a[4]));
    break;
  case 38 : //
    break;
  case 39 : //
    break;
  case 40 : //
    break;
  case 41 : //
    break; 
  case 42 : //
    kqml_critical_section_start();
    break;
  case 43 : //
    kqml_critical_section_end();
    break;
  case 44 : //
    break;
  default:
    TKqml_Exit(interp,TKQML_MOUNT_ERROR,TKQML_STATIC,TCL_ERROR);
  }
  return(OK);
}


static int launch(ClientData clientData, Tcl_Interp *interp, int argc, 
		  char *argv[], void *argt[], TTYPE rc[], int index) {
  int i;
  static char *frame=" ";

  // set return value
  switch (kqml_cmdarray[index].rt) {     // sw. on return type of kril function
  case NUL :                                // void function; nothing to return
    break; 
  case INT :                                   // integer - fold into *int case
  case PINT :                          // set result to integer value of string
    Tcl_SetResult(interp,itoa(*(int*)(argt[0])),TCL_VOLATILE);
    break;
  case CHAR :                                             // return a character
    frame[0]=*(char*)(argt[0]);            // frame character (as small string)
    Tcl_SetResult(interp,frame,TCL_VOLATILE);
    break;
  case PCHAR :                                                   // string case
    Tcl_SetResult(interp,(char*)argt[0],TCL_VOLATILE);
    break;
  case POBJ :                                           // generic pointer case
    Tcl_SetResult(interp,ptoa(argt[0], TKQML_POBJ_TYPE),TCL_VOLATILE);
    break;
  case PPOBJ:
    // always give pointer
    Tcl_SetResult(interp,ptoa(argt[0], TKQML_PPOBJ_TYPE),TCL_VOLATILE);
    break;
  default:                                             // default is do nothing
    break;
  }

  if (argt[1]!=NULL) { free(argt[1]); argt[1]==NULL; }

  // parameters
  for(i=2; i<argc; i++) {
#ifdef DEBUG
    printf("%lu -> rc=%x\n",i-2,rc[i]);
#endif
    if (rc[i] & COPY) {
      switch (rc[i] & RETURN_CODE_TYPE) {
      case KINT : 
#ifdef DEBUG
	  printf("setting %s to %s\n", argv[i],itoa(*((int *)(argt[i]))));
#endif
	  Tcl_SetVar(interp,argv[i],itoa( *((int *)(argt[i])) ),TCL_GLOBAL_ONLY); 
	  break;
      case KCHAR : 
	  Tcl_SetVar(interp,argv[i],(char*)argt[i],TCL_GLOBAL_ONLY); 
	  break;
      case KSTR : 
	  Tcl_SetVar(interp,argv[i],(char*)argt[i],TCL_GLOBAL_ONLY); 
	  break;
      case KPTR : 
	  Tcl_SetVar(interp,argv[i],ptoa(argt[i],TKQML_POBJ_TYPE),
		     TCL_GLOBAL_ONLY); 
	  break;
      case KPPTR : 
	  Tcl_SetVar(interp,argv[i],ptoa(argt[i],TKQML_PPOBJ_TYPE),
		     TCL_GLOBAL_ONLY); 
	  break;
      case KKET:
	  Tcl_SetVar(interp,argv[i],ptoa(argt[i],TKQML_POBJ_TYPE),
		     TCL_GLOBAL_ONLY);
	  break;
      default: 
	  TKqml_Exit(interp,TKQML_TYPE_COPY_ERROR,TKQML_STATIC,TCL_ERROR);
	  break;
      } // switch 2
    }
    if (rc[i] & DESTROY) { free(argt[i]); }
  }
  //Tcl_SetResult(TInterp,rc[0],TCL_GLOBAL_ONLY);
}
//-----------------------------------------------------------------------------
//
//       Miscellaneous utilities
//
//-----------------------------------------------------------------------------

int TKqml_Exit(Tcl_Interp *interp, char *rval, int type, int exit_code) {
  Tcl_SetResult(interp,rval,
		(type==1?TCL_VOLATILE:(type==0?TCL_STATIC:TCL_DYNAMIC)));
  dprint("Application Exiting - %s\n",rval);
  dprint("Application Exiting - %s\n",rval);
  return(exit_code);
}

// upon receiving an error in rval, throw it (return(rval)).
// to flag error, return(tkqml_err(ERROR_NAME))
int tkqml_err(char *rval) {
  dprint("Application Aborting - %s\n",rval);
  Tcl_SetResult(TInterp,rval,TCL_STATIC);
  return(TCL_ERROR);
}

// add varargs
// note: hard-coded to ignore index and add parameters in pairs.
static int add_fields(kqml_message *msg, char *rack[], int index) {
  int i;

  i=0;
  while (rack[i]) {
    kqml_put_field((char*)rack[i],msg,(char*)rack[i+1],
		   strlen((char*)(rack[i+1])));
    i+= 2;
  }
  return(OK);
}





























