//-----------------------------------------------------------------------------
//
//      (3) Performative Handler Code
//
//      The code in this section deals with the handling of received kqml
//      messages.
//
//      register: installs a handler script in the handler table.
//      lookup:   looks up the appropriate script for a performative.
//      master_handler acts as dispatch for the handler scripts.
//
//-----------------------------------------------------------------------------
// $Id: TKqmlHandler.C,v 1.3 1997/05/30 17:51:37 ian Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <kqml.h>
#include <TKqml.h>
#include <TKqmlHandler.h>

char *system_default_handler="TKqmlApp.Default";

// registers (or replaces) all performatives and one default.
//   kqml register ask-if "tcl script"   set/reset script  
//   kqml register ask-if                unset script

int register_handler (ClientData clientData, Tcl_Interp *interp,
	int argc, char* argv[])
{
  int hook;
  char *script;

  if ((hook=HT_LookUp(argv[2]))==KQML_ERR) return (KQML_ERR);   // no such perf
  else {
    if (handler_table[hook].handler!=NULL)                // this is a replace
      free(handler_table[hook].handler);              // so release old script,
    if (argc==2) handler_table[hook].handler=NULL;
    else {
      handler_table[hook].handler=(char*)malloc(strlen(argv[3]));// & store new
      strcpy(handler_table[hook].handler,argv[3]);
      handler_table[hook].handler_is_file = 1;
    }
    return(KQML_OK);
  }
}

int register_handler_as_script (ClientData clientData, Tcl_Interp *interp,
	int argc, char* argv[])
{
  int hook;
  char *script;

  if ((hook=HT_LookUp(argv[2]))==KQML_ERR) return (KQML_ERR);   // no such perf
  else {
    if (handler_table[hook].handler!=NULL)                // this is a replace
      free(handler_table[hook].handler);              // so release old script,
    if (argc==2) handler_table[hook].handler=NULL;
    else {
      handler_table[hook].handler=(char*)malloc(strlen(argv[3]));// & store new
      strcpy(handler_table[hook].handler,argv[3]);
      handler_table[hook].handler_is_file = 0;
    }
    return(KQML_OK);
  }
}

static int HT_LookUp(char *perf) {  // returns an index into the handler table,
  int i;                              // or KQML_ERR if performative not found.

  for (i=0; i<kqml_numhand+1; i++)
    if (strcmp(perf,handler_table[i].perf)==0) return(i);
  return(KQML_ERR);
}

// select, in this order of preference:
// 1. user defined performative handler
// 2. user defined default handler
// 3. system default handler

void *master_handler(char *content, kqml_message *message,
			  KQML_REPLY_TYPE *reply_type)
{
  char *script, *rval, *reply_type_str,*c,*c1,*reply_with;
  int code, rcode;

  rval=Tcl_SetVar(TInterp,"kqml_msg",ptoa(message,TKQML_POBJ_TYPE),
		  TCL_GLOBAL_ONLY);
  rval=Tcl_SetVar(TInterp,"kqml_reply_type",
		  TCL_NULL,TCL_GLOBAL_ONLY);             // prepare vbl in adv.
  rval=Tcl_SetVar(TInterp,"kqml_reply",
		  TCL_NULL,TCL_GLOBAL_ONLY);             // prepare vbl in adv.
  reply_with = kqml_get_field(":reply-with",message,NULL,NULL);
  if ((code=HT_LookUp(kqml_get_performative(message)))!=KQML_ERR) 
    script=handler_table[code].handler;         // look up performative handler
  else if ((code=HT_LookUp("default"))!=KQML_ERR)
    script=handler_table[code].handler;    // if no handler, use default script
  else script=system_default_handler;        // last resort, use system default

  if (script == NULL) 
      script = system_default_handler;

  if (handler_table[code].handler_is_file)                      // run handler
      rcode=Tcl_EvalFile(TInterp, script);
  else
      rcode=Tcl_Eval(TInterp, script);
  if (!reply_with) return NULL;

  if (rcode == TCL_ERROR) {
      dprint ("Error while executing handler script\n");
      dprint ("The last command is : %s\n",TInterp->result); 
      *reply_type = KQML_REPLY_ERROR;
      return TInterp->result;
  }

  reply_type_str=Tcl_GetVar(TInterp,"kqml_reply_type",TCL_GLOBAL_ONLY);
  krt_enum_lookup(reply_type_str,reply_type);
  if (*reply_type==KQML_REPLY_STRING) {
     return(Tcl_GetVar(TInterp,"kqml_reply",TCL_GLOBAL_ONLY));
  }
  else if (*reply_type==KQML_REPLY_KMSG) 
     return((kqml_message*)atop(Tcl_GetVar(TInterp,"kqml_reply",
					   TCL_GLOBAL_ONLY)));  
}














